#!/usr/bin/env python
#
# Author:  Yotam Medini  yotam.medini@gmail.com -- Created: 2009/September/27
# GPL
#

import ctypes
import sys
import time
import types

# local
import values


wait_text = "Waiting for setting receipt... (tap touchpad)"

class AlpsSHM:

    def __init__(self, shm_key, shm_id):
        self.libc = libc = ctypes.CDLL("libc.so.6")
        self.p = None
        if shm_id == None:
            if shm_key == None:
                shm_key = values.SHM_KEY
            shm_id = libc.shmget(shm_key, 0, 0);
        if shm_id >= 0:
            cpi = ctypes.POINTER(ctypes.c_int)
            libc.shmat.restype = cpi
            self.p = libc.shmat(shm_id, 0, 0);
        self.shm_id = shm_id


    def __del__(self):
        if self.p != None:
            sys.stderr.write("Detach p=%s\n" % self.p)
            rc = self.libc.shmdt(self.p)
            sys.stderr.write("shmdt(%s) returned 0x%x\n" % (self.p, rc))


    def ok(self):
        return self.p != None


class GUI:

    def __init__(self, alpsCfg):
        import gobject
        import gtk
        self.gobject = gobject
        self.gtk = gtk
        self.parent = alpsCfg
        self.save_filename = None
        self.spin_buttons = values.SHM_N_VALUES * [None]
        self.sync_map = values.SHM_N_VALUES * [None]
        self.create_sync_map()
        self.info_dlg = None
        self.xy_sync_mode = True
        self.set_count = 0
        self.build()


    def create_sync_map(self):
        for (x, y) in [
            (values.SHM_OPT_xCenterNumerator, values.SHM_OPT_yCenterNumerator),
            (values.SHM_OPT_xCenterDenominator, 
             values.SHM_OPT_yCenterDenominator),
            (values.SHM_OPT_xEdgeNumerator, values.SHM_OPT_yEdgeNumerator),
            (values.SHM_OPT_xEdgeDenominator, 
             values.SHM_OPT_yEdgeDenominator),
            (values.SHM_OPT_xJump, values.SHM_OPT_yJump)]:
            self.sync_map[x] = y
            self.sync_map[y] = x


    def build(self):
        gtk = self.gtk
        self.topwin = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.topwin.set_size_request(860, 500)
        self.topwin.connect("delete_event", self.delete_event, None)
        self.topwin.connect("destroy", self.destroy, None)
        self.topwin.set_border_width(10)
        self.topwin.set_title("ALPS - Touch-Pad Configuration")
        self.topwin.connect('delete-event', gtk.main_quit)
        self.tooltips = gtk.Tooltips()
        vbox_main = gtk.VBox(False, 0)
        vbox = gtk.VBox(False, 0)
        bars = self.build_top_bars()
        for bar in bars:
            vbox_main.pack_start(bar, False, False, 0)
        tbl = self.build_table()
        f = gtk.Frame()
        f.add(tbl)
        vbox.pack_start(f, True, True, 2)
        vbox.pack_start(self.create_emulate_middle(), True, True, 2)
        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        sw.add_with_viewport(vbox)
        vbox_main.pack_start(sw, True, True, 1)
        self.statusbar = gtk.Statusbar()
        self.sb_ctxid = self.statusbar.get_context_id("general")
        vbox_main.pack_start(self.statusbar, False, False, 0)
        self.topwin.add(vbox_main)
        
        self.topwin.show_all()


    def build_top_bars(self):
        gtk = self.gtk
        ui_info = '''
<ui>
  <menubar name='MenuBar'>
    <menu action='FileMenu'>
      <menuitem action='Apply'/>
      <menuitem action='Save'/>
      <menuitem action='SaveAs'/>
      <separator/>
      <menuitem action='Quit'/>
    </menu>
    <menu action='PreferencesMenu'>
      <menu action='Sync'>
        <menuitem action='XYSynchronize'/>
        <menuitem action='XYSeparate'/>
      </menu>
    </menu>
    <menu action='HelpMenu'>
      <menuitem action='About'/>
    </menu>
  </menubar>
  <toolbar  name='ToolBar'>
    <toolitem action='Apply'/>
    <toolitem action='Show'/>
    <toolitem action='Quit'/>
  </toolbar>
</ui>'''

        entries = (  # name, stock id, label
            ( "FileMenu", None, "_File" ), 
            ( "PreferencesMenu", None, "_Preferences" ),
            ( "Sync", gtk.STOCK_DISCONNECT, "_Sync" ),
            ( "HelpMenu", None, "_Help" ),
            ( "Apply", gtk.STOCK_APPLY,
              "_Apply","<control>A",
              "Apply settings",
              self.apply),
            ( "Save", gtk.STOCK_SAVE,
              "_Save","<control>S",
              "Save current file",
              self.save),
            ( "SaveAs", gtk.STOCK_SAVE,
              "Save _As...", "<control><shift>S",
              "Save to a file",
              self.save_as ),
            ( "Quit", gtk.STOCK_QUIT,
              "_Quit", "<control>Q",
              "Quit",
              self.quit ),
            ( "Show", gtk.STOCK_PRINT_REPORT,
              "Show", "<control>P",
              "print values in hal.fdi format to standard output",
              self.show ),
            ( "About", gtk.STOCK_ABOUT,
              "_About", "<control>A",
              "About",
              self.about ),
        )

        sync_entries = (
            ( "XYSynchronize", gtk.STOCK_CONNECT,
              "XY-S_ynchronize", "<control>Y",
              "Synchronize ratios and jump values for X & Y", True ),
            ( "XYSeparate", gtk.STOCK_DISCONNECT,
              "XY-Se_parate", "<control>P",
              "Separate ratios and jump values for X & Y", False ),
        )
        actions = gtk.ActionGroup("Actions")
        actions.add_actions(entries)
        actions.add_radio_actions(sync_entries, True, self.xy_radio_sync)
        ui = gtk.UIManager()
        ui.insert_action_group(actions, 0)
        self.topwin.add_accel_group(ui.get_accel_group())

        try:
            mergeid = ui.add_ui_from_string(ui_info)
        except gobject.GError, msg:
            print "building menus failed: %s" % msg

        mb = ui.get_widget("/MenuBar")
        tb = ui.get_widget("/ToolBar")
        tb.set_tooltips(True)
        return [mb, tb]


    def build_table(self):
        gtk = self.gtk
        (nc, nr) = (9, 6)
        tbl = gtk.Table(nc, nr, False)

        for (label, c, r, shmi, max_value) in [
            ("xMin", 0,      nr/2,           values.SHM_OPT_xPadMin,  300),
            ("xMax", nc - 1, nr/2,           values.SHM_OPT_xPadMax, 2000),
            ("yMin", nc/2,   0,              values.SHM_OPT_yPadMin,  300),
            ("yMax", nc/2,   nr - 1,         values.SHM_OPT_yPadMax, 2000),
            ("xJump", nc/4,  nr - 2,         values.SHM_OPT_xJump, 200),
            ("yJump", nc - 1 - nc/4, nr - 2, values.SHM_OPT_yJump, 200)]:
            vertical = label[0] == 'x' or label[1:] == 'Jump'
            w = self.create_numeric(label, shmi, max_value, vertical)
            self.table_put(tbl, w, c, r)

        for (label, c, r, shm_ni, shm_di) in [
            ("xEdge", 1, 1, values.SHM_OPT_xEdgeNumerator, 
             values.SHM_OPT_xEdgeDenominator),
            ("xCenter",  nc/2 - 1, nr/2, values.SHM_OPT_xCenterNumerator, 
             values.SHM_OPT_xCenterDenominator),
            ("yEdge", nc - 2, 1, values.SHM_OPT_yEdgeNumerator,
             values.SHM_OPT_yEdgeDenominator),
            ("yCenter", nc/2 + 1, nr/2, values.SHM_OPT_yCenterNumerator,
             values.SHM_OPT_yCenterDenominator)]:
            w = self.create_ratio(label, shm_ni, shm_di)
            self.table_put(tbl, w, c, r)
         
        return tbl


    def create_emulate_middle(self):
        gtk = self.gtk
        w = self.create_numeric("Emulate Middle Button", 
                                values.SHM_OPT_EmulateMidButtonTime, 1000, 
                                False)
        return self.hbox_put_middle(w)



    def table_put(self, table, child, c, r):
        table.attach(child, c, c + 1, r, r + 1, xpadding=2, ypadding=2)


    def create_numeric(self, label, shmi, max_value, vertical):
        gtk = self.gtk
        box = gtk.VBox() if vertical else gtk.HBox()
        box.pack_start(gtk.Label(label), False, False, 2)
        adj = gtk.Adjustment(1, 0, max_value, 1)
        sb = self.create_spin_button_shmi(adj, shmi)
        sb.connect("value-changed", self.value_changed, shmi)
        if vertical:
            box.pack_start(self.hbox_put_middle(sb))
        else:
            box.pack_start(sb, False, False, 2)
        return box
        

    def create_ratio(self, label, shm_ni, shm_di):
        gtk = self.gtk
        tbl = gtk.Table(2,3)
        tbl.attach(gtk.Label(label), 0, 1, 0, 3, xpadding=1)

        names = []
        for (shmi, r) in [(shm_ni, 0), (shm_di, 2)]:
            names.append(values.names[shmi])
            adj = gtk.Adjustment(1, 1, 16, 1)
            sb = self.create_spin_button_shmi(adj, shmi)
            sb.connect("value-changed", self.value_changed, shmi)
            sb.set_width_chars(2)
            self.table_put(tbl, sb, 1, r)
        self.table_put(tbl, gtk.HSeparator(), 1, 1)

        self.tooltips.set_tip(tbl, """
Set ratio:
  %s / %s
  <screen-pixel> / <pad-points>.
Bigger ratio makes "faster" move
"""[1:] % tuple(names))

        return tbl


    def create_spin_button_shmi(self, adj, shmi):
        gtk = self.gtk
        sb = gtk.SpinButton(adj)
        v = self.parent.get_shm_option(shmi)
        sb.set_value(v)
        self.spin_buttons[shmi] = sb
        self.tooltips.set_tip(sb, "Set %s" % values.names[shmi])
        return sb


    def hbox_put_middle(self, w):
        gtk = self.gtk
        h = gtk.HBox()
        h.pack_start(gtk.Label(""), True, True, 2)
        h.pack_start(w, False, False, 2)
        h.pack_start(gtk.Label(""), True, True, 2)
        return h


    def value_changed(self, sb, shmi):
        self.status("Changes not yet applied")
        imhs = self.sync_map[shmi]
        if self.xy_sync_mode and imhs != None:
            self.spin_buttons[imhs].set_value(sb.get_value_as_int())


    def apply(self, dum):
        change = False
        for i in range(0, values.SHM_N_VALUES):
            sb = self.spin_buttons[i]
            v = None if sb == None else sb.get_value_as_int()
            if v != None and self.parent.set_shm_option(i, v):
                change = True
        if change:
            self.info(wait_text)
            self.parent.wait_shm_cmd()
            self.info(None)
            self.set_count += 1
            self.status("%d setting(s) done" % self.set_count)
        else:
            self.warning("No change\n")


    def status(self, text):
        self.statusbar.pop(self.sb_ctxid)
        self.statusbar.push(self.sb_ctxid, text)


    def save(self, dum=None):
        if self.save_filename == None:
            self.save_as()
        else:
            self.save_fn(self.save_filename)


    def save_as(self, dum=None):
        gtk = self.gtk
        dialog = gtk.FileChooserDialog("Select .fdi file",
                                       self.topwin,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                        gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            self.save_filename = dialog.get_filename()
            self.save(self.save_filename)
        dialog.destroy()


    def save_fn(self, fn):
        self.parent.fdi_save(fn)

    def show(self, dum):
        self.parent.hal_write(sys.stdout)


    def quit(self, dum):
        self.gtk.main_quit()


    def xy_radio_sync(self, action, current):
        self.xy_sync_mode = current.get_current_value()


    def about(self, action):
        dialog = self.gtk.AboutDialog()
        dialog.set_name("ALPS Touch-Pad Configuration")
        dialog.set_copyright("\302\251 Copyright 2009 Yotam Medini")
        dialog.set_website("http://www.x.org./")
        dialog.set_authors(["Yotam Medini <yotam DOT medini AT gmail DOT ...>"])
        dialog.set_version("0.1")
        ## Close dialog on user response
        dialog.connect ("response", lambda d, r: d.destroy())
        dialog.show()


    def warning(self, text):
        gtk = self.gtk
        w = gtk.MessageDialog(self.topwin, gtk.DIALOG_MODAL, 
                              type=gtk.MESSAGE_WARNING,
                              buttons=gtk.BUTTONS_CANCEL,
                              message_format=text)
        w.run()
        w.destroy()


    def info(self, text):
        gtk = self.gtk
        if self.info_dlg != None:
            self.info_dlg.destroy()
            self.info_dlg = None
        if text != None:
            w = gtk.MessageDialog(self.topwin, flags=gtk.DIALOG_MODAL, 
                                  type=gtk.MESSAGE_WARNING,
                                  buttons=gtk.BUTTONS_NONE,
                                  message_format=text)
            self.info_dlg = w
            w.show_all()
            self.gtk_flush()
            time.sleep(0.1)


    def gtk_flush(self):
        gtk = self.gtk
        while gtk.events_pending():
            gtk.main_iteration()


    def delete_event(self, widget, event, data=None):
        return False

    def destroy(self, widget, event, data=None):
        self.gtk.main_quit()

    def run(self):
        self.gtk.main()


class AlpsCfg:

    def usage(self):
        self.helped = True
        sys.stderr.write("""
Usage:
%s
  [-h | --help]         # This message
  [-q | --query]
  [-qhal]               # Query printed in hal.fdi format
  [-gui]
                        # Set Values:
  [-tmb <milliseconds>]   # EmulateMidButtonTime
  [-xmm <min max>]        # xPadMin, xPadMax
  [-ymm <min max>]        # yPadMin, yPadMax
  [-xcnd <n d>]           # xCenter: Numerator Denominator
  [-ycnd <n d>]           # yCenter: Numerator Denominator
  [-xend <n d>]           # xEdge: Numerator Denominator
  [-yend <n d>]           # yEdge: Numerator Denominator
  [-xj <n>]               # xJump
  [-yj <n>]               # yJump

  [-shmkey <key>]       # Default: %d
  [-shmid <id>]         # Default: via shmkey
"""[1:] % (self.argv[0], values.SHM_KEY))


    def __init__(self, argv):
        self.rc = 0
        self.helped = False
        self.argv = argv
        ai = 1
        self.set_opt2val = so2v = {}
        self.query = False
        self.qhal = False
        self.gui = False
        self.g = None
        self.shm = None
        self.shmkey = None
        self.shmid = None
        while self.may_run() and ai < len(argv):
            opt = argv[ai]
            ai += 1
            if opt in ('-h', '--h', '-help', '--help'):
                self.usage()
            elif opt in ('-q', '-query', '--query'):
                self.query = True
            elif opt == '-qhal':
                self.qhal = True
            elif opt == '-gui':
                self.gui = True
            elif opt == '-tmb':
                so2v[values.SHM_OPT_EmulateMidButtonTime] = int(argv[ai])
                ai += 1
            elif opt in ('-xmm', '-ymm'):
                m = int(argv[ai]);  ai += 1
                M = int(argv[ai]);  ai += 1
                if opt == '-xmm':
                    so2v[values.SHM_OPT_xPadMin] = m
                    so2v[values.SHM_OPT_xPadMax] = M
                else:
                    so2v[values.SHM_OPT_yPadMin] = m
                    so2v[values.SHM_OPT_yPadMax] = M
            elif opt in ('-xcnd', '-ycnd', '-xend', '-yend'):
                n = int(argv[ai]);  ai += 1
                d = int(argv[ai]);  ai += 1
                if opt == '-xcnd':
                    so2v[values.SHM_OPT_xCenterNumerator] = n
                    so2v[values.SHM_OPT_xCenterDenominator] = d
                elif opt == '-ycnd':
                    so2v[values.SHM_OPT_yCenterNumerator] = n
                    so2v[values.SHM_OPT_yCenterDenominator] = d
                elif opt == '-xend':
                    so2v[values.SHM_OPT_xEdgeNumerator] = n
                    so2v[values.SHM_OPT_xEdgeDenominator] = d
                else:
                    so2v[values.SHM_OPT_yEdgeNumerator] = n
                    so2v[values.SHM_OPT_yEdgeDenominator] = d
            elif opt in ('-xj', '-yj'):
                j = int(argv[ai]);  ai += 1
                if opt == '-xj':
                    so2v[values.SHM_OPT_xJump] = j
                else:
                    so2v[values.SHM_OPT_yJump] = j
            elif opt == '-shmkey':
                self.shmkey = int(argv[ai]);  ai += 1
            elif opt == '-shmid':
                self.shmid = int(argv[ai]);  ai += 1
            else:
                self.err("Unexpected option: '%s'" % opt)


    def __del__(self):
        self.del_shm()


    def del_shm(self):
        if self.shm != None:
            del self.shm
            self.shm = None


    def ok(self):
        return self.rc == 0


    def may_run(self):
        return self.ok() and not self.helped


    def err(self, msg):
        if self.ok():
            sys.stderr.write(msg + "\n")
            self.rc = 1
            self.usage()


    def message(self, msg):
        if msg != None:
            sys.stderr.write(msg + "\n")


    def run(self):
        if False: # TEMPORARY YYYYYYYYYYYYYYYYYYYYYYYYYYYYYYYY
            if self.gui:
                g = GUI(self)
                g.run()
                sys.exit(13)
        self.shm = AlpsSHM(self.shmkey, self.shmid)
        if not self.shm.ok():
            self.err("Failed to attach shared memory")
        else:
            if self.query:
                self.show()
            if self.qhal:
                self.hal_write(sys.stdout)
            self.set_cmdline_opts()
            if self.gui:
                g = GUI(self)
                g.run()


    def show(self):
        for i in range(0, values.SHM_N_VALUES):
            sys.stdout.write("%-20s = %3d\n" %
                             (values.names[i], self.shm.p[i + 1]))


    def fdi_save(self, fn):
        f = open(fn, 'w')
        f.write(values.head_fdi)
        self.hal_write(f)
        f.write(values.tail_fdi)


    # Following methods that access shm, add +1 to skip 
    # SharedMemConfig.MailBox.cmd

    def hal_write(self, f):
        indent = 8*' '
        pathdir = "input.x11_options"
        for i in range(0, values.SHM_N_VALUES):
            name = values.names[i]
            v = self.get_shm_option(i)
            f.write('%s<merge key="%s.%s" type="string">%d</merge>\n' %
                    (indent, pathdir, name, v))


    def set_cmdline_opts(self):
        change = False
        for i in range(0, values.SHM_N_VALUES):
            v = self.set_opt2val.get(i, None)
            if v != None:
                if self.set_shm_option(i, v):
                    change = True
        if change:
            self.message(wait_text)
            self.wait_shm_cmd()
            self.message("Setting received")
        elif self.set_opt2val:
            sys.stderr.write("No changed values\n")


    def wait_shm_cmd(self):
        self.shm.p[0] = 1
        nap = 0.1
        while self.shm.p[0] != 0:
            time.sleep(nap)
            nap = min(2*nap, 1.)


    def get_shm_option(self, shmi):
        return self.shm.p[shmi + 1]


    def set_shm_option(self, shmi, v):
        change = self.shm.p[shmi + 1] != v
        if change:
            change = self.shm.p[shmi + 1] = v
        return change


p = AlpsCfg(sys.argv)
if p.may_run():
    p.run()
rc = p.rc
sys.stderr.write("after run: rc=%d\n" % rc)
p.del_shm()
sys.exit(rc)
